use std::fs;
use std::io::ErrorKind;
use std::path::Path;
use std::rc::Rc;

use crate::db::*;
use crate::ui::cards::new::*;
use crate::widgets::card_entry::CardEntry;
use glib::subclass::InitializingObject;
use gtk::glib::object::ObjectExt;
use gtk::glib::{clone, closure_local};
use gtk::subclass::prelude::*;
use gtk::{gio, glib, Box, Button, CompositeTemplate, ScrolledWindow, SearchEntry, Window};
use gtk::{prelude::*, FileDialog};
use rusqlite::Connection;
use sha256::try_digest;

#[derive(CompositeTemplate, Default)]
#[template(resource = "/org/foxarmy/learn-hieroglyph/cards/edit/ui.xml")]
pub struct MemoryCardsEditScene {
    #[template_child]
    pub search_entry: TemplateChild<SearchEntry>,
    #[template_child]
    pub add_button: TemplateChild<Button>,
    #[template_child]
    pub cards: TemplateChild<Box>,
    #[template_child]
    pub cards_scrolled_window: TemplateChild<ScrolledWindow>,
}

#[glib::object_subclass]
impl ObjectSubclass for MemoryCardsEditScene {
    const NAME: &'static str = "MemoryCardsEditScene";
    type Type = super::MemoryCardsEditScene;
    type ParentType = gtk::ApplicationWindow;

    fn class_init(klass: &mut Self::Class) {
        klass.bind_template();
    }

    fn instance_init(obj: &InitializingObject<Self>) {
        obj.init_template();
    }
}

impl ObjectImpl for MemoryCardsEditScene {
    fn constructed(&self) {
        self.parent_constructed();

        let binding = self.obj();

        self.add_button.connect_closure("clicked", 
                                        false,
                                        closure_local!(@strong binding => move |_b: &Button| {
                                        let new_win = Rc::new(MemoryCardsNewScene::new(&binding.application().unwrap()));
                                        new_win.get_file_choose_button().connect_clicked(clone!(@strong new_win => move |_| {
                                            gtk::glib::MainContext::default().spawn_local(file_choose_dialog(Rc::clone(&new_win)));
                                        }));
                                        new_win.present();
        }));
        self.update_card_list();
    }

}

impl MemoryCardsEditScene {
    pub fn update_card_list(&self) {
        let c: &Box = self.cards.as_ref();

        while c.first_child() != None {
            c.remove(&c.first_child().unwrap());
        }
    
        let conn = Connection::open(get_db_path()).unwrap();
        let mut stmt = conn
            .prepare("SELECT imagename, hieroglyph, reading FROM cards")
            .unwrap();
        let cards_iter = stmt
            .query_map([], |row| {
                let a: String = row.get(0).unwrap();
                let b: String = row.get(1).unwrap();
                let c: String = row.get(2).unwrap();
                let entry = CardEntry::new(&a, &b, &c);
                Ok(entry)
            })
            .unwrap();
    
        for card in cards_iter {
            let card_binding = card.unwrap();
            let path = get_images_store_path() +  &card_binding.imagepath();
            card_binding.set_imagepath(&*path);
            card_binding.get_picture_widget().set_file(Some(&gio::File::for_path(&path)));
            c.append(&card_binding);
        }
    }
}



async fn file_choose_dialog<W: IsA<gtk::Window>>(window: Rc<W>) {
    let dialog: FileDialog = gtk::FileDialog::builder().build();
    let answer = dialog.open_future(Some(&*window)).await;
    let path = answer.unwrap().path().unwrap();
    let path: &str = path.as_path().to_str().unwrap();
    let w: &MemoryCardsNewScene = Into::<&Window>::into(window.upcast_ref())
        .downcast_ref()
        .unwrap(); // Weird casting from &Window as passed in func to &MemoryCardsNewScene

    w.get_picture_widget()
        .set_file(Some(&gio::File::for_path(path)));

    let images_store_path = get_program_home_path() + "/images";

    match fs::create_dir_all(&images_store_path) {
        Ok(_) => {}
        Err(error) => match error.kind() {
            ErrorKind::AlreadyExists => {}
            _ => panic!("Could not create directory for storing pictures!"),
        },
    };

    let hash = try_digest(path).unwrap();
    let extenstion = Path::new(path).extension().unwrap().to_str().unwrap();
    let new_filename: String = hash.as_str().to_owned() + "." + extenstion;
    let stored_image_path = Path::new(&images_store_path).join(&new_filename);
    fs::copy(path, stored_image_path).expect("Error copying image to store");

    w.get_done_button().connect_closure(
        "clicked",
        false,
        closure_local!(@strong w => move |_b: &Button| {
            let hieroglyph = w.get_hieroglyph_input();
            let reading = w.get_reading_input();
            let conn = Connection::open(get_db_path()).unwrap();

            let query = "INSERT INTO cards (imagename, hieroglyph, reading) VALUES(?1, ?2, ?3)";
            conn.execute(query, (&new_filename, &hieroglyph, &reading)).unwrap();
            w.close();

        }),
    );
}

impl WidgetImpl for MemoryCardsEditScene {}

impl WindowImpl for MemoryCardsEditScene {}

impl ApplicationWindowImpl for MemoryCardsEditScene {}
